VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "CameraFPS"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
  '*************************************************************
  'Name: CameraFPS
  'Author: Jeff Weber
  'Date: 2/13/2002
  'Description:
  'This class provides first person shooter movement for a
  'TrueVision3D.TVCamera objects.  The rotations are done with matrix
  'math rather than keeping track of Sins and Cosines.
  'Features:
  ' :Smoothing Factor:
  '  "coast" to a stop rather than stopping abruptly.
  '
  ' :Jumping:
  '  Allows you to jump up and over things.
  '
  ' :Gravity:
  '   Makes you fall down and affects how high you can jump.
  '
  '*************************************************************
  
  Option Explicit
  '--Internal TrueVision3D.TVCamera object.
  Private mCamera           As TrueVision3D.TVCamera
  '--Landscape object. This gets a reference to the
  '  users landscape object at Initialization time.
  Private mLandscape        As TrueVision3D.TVLandscape
  
  '***************************************************************
  'These parameters are used to control camera rotation and tilt.
  '***************************************************************
  Private mmatRotY          As D3DMATRIX
  Private mmatPhantomCam    As D3DMATRIX
  Private mLookAtHeight     As Single
  Private mMaxLookAtHeight  As Single
  Private mMinLookAtHeight  As Single
    
  '***************************************************************
  'These are the primary parameters for controling the state
  'of the camera.
  '***************************************************************
  Private msngRaiseLower    As Single
  Private msngStrafe        As Single
  Private msngFrontBack     As Single
  Private msngRotation      As Single
  Private msngTilt          As Single
  Private msngRestHeight        As Single
  Private mvecPosition          As D3DVECTOR
  Private mvecLookAt            As D3DVECTOR
  Private mvecFront             As D3DVECTOR
  Private mvecSide              As D3DVECTOR
    
  '**************************************************************
  'These params control the motion of the camera
  '**************************************************************
  Private mblnRun           As Boolean
  Private msngMoveSpeed     As Single
  Private msngWalkSpeed     As Single
  Private msngRunSpeed      As Single
  Private msngSmoothing     As Single
  
  '**************************************************************
  'These params are used to control jumping.
  '**************************************************************
  Private msngGravity           As Single
  Private msngYPosition         As Single
  Private msngYVelocity         As Single
  Private msngJumpForce         As Single
  Private msngRestAltitude      As Single
  Private mblnIsJumping         As Boolean
  Private msngTempFB            As Single
  Private msngTempS             As Single
  
  Private msngTimeElapsed       As Single
  
  Private Const LOOK_AT_DISTANCE  As Single = 400
  '--Epsilon contansts to prevent having to deal with
  '  very small numbers.
  Private Const EPSILON_GRAVITY   As Single = 2
  Private Const EPSILON_SMOOTHING As Single = 0.0001
  
  '*************************************************************
  'These constants are simply used to kinda normalize the
  'different settable camera variables. So instead of setting
  'gravity = -.01, you can use gravity = -1. I used trial and
  'error to get what I thought were decent values.
  '*************************************************************
  Private Const NORMALIZE_FACTOR_ROTATION As Single = 1 / 2000
  Private Const NORMALIZE_FACTOR_TILT     As Single = 1 / 2
  Private Const NORMALIZE_FACTOR_MOVE     As Single = 1 / 10
  Private Const NORMALIZE_FACTOR_GRAVITY  As Single = 1 / 10
  
  Private Const ERR_MAX_TILT As Long = vbObjectError + 1000
  Private Const ERR_MIN_TILT As Long = vbObjectError + 1001
  
  
  '******************************************************************
  'Intitialize the camera.  A reference to the user landscape object
  'must be passed in because it is needed to control camera height.
  '******************************************************************
  Public Function Initialize(ByVal Xpos As Single, _
                             ByVal ZPos As Single, _
                             ByVal RestHeight As Single, _
                             Landscape As TrueVision3D.TVLandscape)
    
    Set mLandscape = Landscape
    '--Default RestHeight
    msngRestHeight = RestHeight
    '--Set cameras initial position
    mvecPosition = Vector3(Xpos, mLandscape.GetHeight(Xpos, ZPos) + msngRestHeight, ZPos)
    With mvecPosition
      mCamera.SetPosition .x, .y, .z
      '--This variable is used for jumping and needs to be initialized here.
      msngYPosition = .y
    End With
    
    '--Walk by default
    mblnRun = False
    '--Use a little trig to set up the MaxCameraTilt and MinCameraTilt.
    '  The default Max, Min tilt is set to 80 but can bet set to
    '  anything between 0 and 90, 0 and -90 respectively.
    mMaxLookAtHeight = LOOK_AT_DISTANCE * Sin(Deg2Rad(80))
    mMinLookAtHeight = -LOOK_AT_DISTANCE * Sin(Deg2Rad(Abs(-80)))
    msngWalkSpeed = 1 * NORMALIZE_FACTOR_MOVE
    msngRunSpeed = 3 * NORMALIZE_FACTOR_MOVE
    msngMoveSpeed = msngWalkSpeed
    msngSmoothing = 200
    msngGravity = -1 * NORMALIZE_FACTOR_GRAVITY
  End Function
  
  Public Property Let Run(Value As Boolean)
    If Value Then
      msngMoveSpeed = msngRunSpeed
    Else
      msngMoveSpeed = msngWalkSpeed
    End If
    mblnRun = Value
  End Property
  
  Public Property Get Run() As Boolean
    Run = mblnRun
  End Property
  
  Public Property Get Camera() As TrueVision3D.TVCamera
    Set Camera = mCamera
  End Property
  
  Public Sub SetWalkSpeed(Speed As Single)
    msngWalkSpeed = Speed * NORMALIZE_FACTOR_MOVE
  End Sub
  
  Public Sub SetRunSpeed(Speed As Single)
    msngRunSpeed = Speed * NORMALIZE_FACTOR_MOVE
  End Sub
  
  Public Sub SetPosition2D(Xpos As Single, ZPos As Single)
    With mvecPosition
      .x = Xpos
      .y = mLandscape.GetHeight(Xpos, ZPos) + msngRestHeight
      .z = ZPos
    End With
  End Sub
  
  Public Sub SetPosition3D(Xpos As Single, YPos As Single, ZPos As Single)
    With mvecPosition
      .x = Xpos
      .y = YPos
      .z = ZPos
    End With
    msngYPosition = YPos
  End Sub
  
  Public Function GetPosition() As D3DVECTOR
    GetPosition = mCamera.GetPosition
  End Function
  
  Public Sub SetGravity(Gravity As Single)
    msngGravity = Gravity * NORMALIZE_FACTOR_GRAVITY
  End Sub
  
  Public Sub SetMaxMinTilt(MaxTiltDegrees As Single, MinTiltDegrees As Single)
    '--Defensive error trapping. This will force the developer to use
    '  valid parameters and prevent run time anomolies.
    On Error GoTo ERR_HANDLER
    If MaxTiltDegrees > 90 Or MaxTiltDegrees < 0 Then
      Err.Raise ERR_MAX_TILT
    End If
    If MinTiltDegrees < -90 Or MinTiltDegrees > 0 Then
      Err.Raise ERR_MIN_TILT
    End If
    
    '--Update the variables that control tilt.
    mMaxLookAtHeight = LOOK_AT_DISTANCE * Sin(Deg2Rad(MaxTiltDegrees))
    mMinLookAtHeight = -LOOK_AT_DISTANCE * Sin(Deg2Rad(Abs(-MaxTiltDegrees)))
    
ERR_HANDLER:
    Select Case Err.Number
      Case ERR_MAX_TILT
        Err.Raise Err.Number, "CameraFPS.SetMaxMinTilt", "CameraFPS.SetMaxMinTilt" & vbCrLf & vbCrLf & "Max tilt angle must be between 0 and 90"
      Case ERR_MIN_TILT
        Err.Raise Err.Number, "", "CameraFPS.SetMaxMinTilt" & vbCrLf & vbCrLf & "Min tilt angle must be between 0 and -90"
    End Select
  End Sub
  
  Public Sub Raise()
    msngRaiseLower = msngMoveSpeed
  End Sub

  Public Sub Lower()
    msngRaiseLower = -msngMoveSpeed
  End Sub

  Public Sub MoveLeft()
    If Not (mblnIsJumping) Then
      msngStrafe = -msngMoveSpeed
    End If
  End Sub

  Public Sub MoveRight()
    If Not (mblnIsJumping) Then
      msngStrafe = msngMoveSpeed
    End If
  End Sub

  Public Sub MoveForward()
    If Not (mblnIsJumping) Then
      msngFrontBack = msngMoveSpeed
    End If
  End Sub

  Public Sub MoveBackward()
    If Not (mblnIsJumping) Then
      msngFrontBack = -msngMoveSpeed
    End If
  End Sub

  Public Sub TiltUpDown(ByRef Speed As Single)
    msngTilt = Speed * NORMALIZE_FACTOR_TILT
  End Sub
  
  Public Sub LeanLeft()
    
  End Sub

  Public Sub Rotate(ByRef Speed As Single)
    msngRotation = msngRotation + Speed * NORMALIZE_FACTOR_ROTATION
  End Sub

   Public Sub Jump(Force As Single)
    'If mvecPosition.Y < msngRestAltitude + EPSILON_GRAVITY Then
    If mblnIsJumping = False Then
      msngYVelocity = Force
      msngTempFB = msngFrontBack
      msngTempS = msngStrafe
      mblnIsJumping = True
    End If
  End Sub
  
  Public Sub Update(TimeElapsed As Single)
    msngTimeElapsed = TimeElapsed
    CalculateRotation
    CalculatePosition
    CalculateLookAt
    With mvecPosition
      msngRestAltitude = mLandscape.GetHeight(.x, .z) + msngRestHeight
      If .y < msngRestAltitude Then
        .y = msngRestAltitude
      End If
      CalculateGravity
      mCamera.SetPosition .x, .y, .z
    End With
    mCamera.SetLookAt mvecLookAt.x, mvecLookAt.y, mvecLookAt.z
    msngRotation = 0
  End Sub
  
  Private Sub CalculateGravity()
    With mvecPosition
      If .y > msngRestAltitude + EPSILON_GRAVITY Or msngYVelocity > 0 Then
        msngYVelocity = msngYVelocity + msngGravity
        msngYPosition = msngYPosition + msngYVelocity '* msngTimeElapsed / 10
        .y = msngYPosition
        msngFrontBack = msngTempFB
        msngStrafe = msngTempS
      Else
        If msngYVelocity <= 0 Then
          mblnIsJumping = False
          msngTempFB = 0
          msngTempS = 0
          msngYPosition = msngRestAltitude
          .y = msngRestAltitude
          msngYVelocity = 0
        End If
      End If
    End With
  End Sub
  
  Private Sub CalculateRotation()
'    If msngRotation > 0 Then
'      msngRotation = msngRotation - (msngRotation / msngSmoothing * msngTimeElapsed)
'      If msngRotation < EPSILON_SMOOTHING Then msngRotation = 0
'    End If
'    If msngRotation < 0 Then
'      msngRotation = msngRotation - (msngRotation / msngSmoothing * msngTimeElapsed)
'      If msngRotation > -EPSILON_SMOOTHING Then msngRotation = 0
'    End If
'
    D3DXMatrixRotationY mmatRotY, msngRotation * msngTimeElapsed 'possible smoothing here.
    mCamera.RotateY msngRotation
    With mmatPhantomCam
      .m41 = 0
      .m42 = 0
      .m43 = 0
    End With
    D3DXMatrixMultiply mmatPhantomCam, mmatPhantomCam, mmatRotY
    With mmatPhantomCam
      .m41 = mvecPosition.x
      .m42 = mvecPosition.y
      .m43 = mvecPosition.z
    End With
  End Sub
  
  Private Sub CalculatePosition()
    Dim vecSum   As D3DVECTOR
    Dim vecFront As D3DVECTOR
    Dim vecSide As D3DVECTOR
    
    If msngSmoothing < 1 Then msngSmoothing = 1
       
    If msngFrontBack > 0 Then
      msngFrontBack = msngFrontBack - (msngFrontBack / msngSmoothing * msngTimeElapsed)
      If msngFrontBack < EPSILON_SMOOTHING Then msngFrontBack = 0
    End If
    If msngFrontBack < 0 Then
      msngFrontBack = msngFrontBack - (msngFrontBack / msngSmoothing * msngTimeElapsed)
      If msngFrontBack > -EPSILON_SMOOTHING Then msngFrontBack = 0
    End If
    
    If msngStrafe > 0 Then
      msngStrafe = msngStrafe - (msngStrafe / msngSmoothing * msngTimeElapsed)
      If msngStrafe < EPSILON_SMOOTHING Then msngStrafe = 0
    End If
    If msngStrafe < 0 Then
      msngStrafe = msngStrafe - (msngStrafe / msngSmoothing * msngTimeElapsed)
      If msngStrafe > -EPSILON_SMOOTHING Then msngStrafe = 0
    End If
    
    If msngRaiseLower > 0 Then
      msngRaiseLower = msngRaiseLower - (msngRaiseLower / msngSmoothing * msngTimeElapsed)
      If msngRaiseLower < EPSILON_SMOOTHING Then msngRaiseLower = 0
    End If
    If msngRaiseLower < 0 Then
      msngRaiseLower = msngRaiseLower - (msngRaiseLower / msngSmoothing * msngTimeElapsed)
      If msngRaiseLower > -EPSILON_SMOOTHING Then msngRaiseLower = 0
    End If
    
    msngRestHeight = msngRestHeight + msngRaiseLower
    
    If Not (mblnIsJumping) Then
      mvecFront = GetFront
      mvecSide = GetSide
    End If
    
    D3DXVec3Scale vecFront, mvecFront, msngFrontBack * msngTimeElapsed
    D3DXVec3Scale vecSide, mvecSide, msngStrafe * msngTimeElapsed
    D3DXVec3Add vecSum, vecFront, vecSide
    D3DXVec3Add mvecPosition, mvecPosition, vecSum
  End Sub
   
  Private Sub CalculateLookAt()
    mLookAtHeight = mLookAtHeight + msngTilt
    If mLookAtHeight > mMaxLookAtHeight Then mLookAtHeight = mMaxLookAtHeight
    If mLookAtHeight < mMinLookAtHeight Then mLookAtHeight = mMinLookAtHeight
    mvecLookAt = GetWorldPosition(mmatPhantomCam, mvecPosition, Vector3(0, mLookAtHeight, LOOK_AT_DISTANCE))
  End Sub
    
  Public Function GetFront() As D3DVECTOR
    With mmatPhantomCam
      GetFront = Vector3(.m31, .m32, .m33)
    End With
  End Function
  
  Public Function GetSide() As D3DVECTOR
    With mmatPhantomCam
      GetSide = Vector3(.m11, .m12, .m13)
    End With
  End Function
  
  Private Sub Class_Initialize()
    Set mCamera = New TrueVision3D.TVCamera
    D3DXMatrixIdentity mmatPhantomCam
  End Sub
  
  Private Sub Class_Terminate()
     Set mCamera = Nothing
  End Sub

'***************************************************************************************
'Misc Math Utility Functions
'***************************************************************************************
Public Function GetWorldPosition(Orientation As D3DMATRIX, Position As D3DVECTOR, LocalVector As D3DVECTOR) As D3DVECTOR
  Dim vecReturn As D3DVECTOR
  Dim matRot As D3DMATRIX
  matRot = Orientation
  With matRot
    .m41 = Position.x
    .m42 = Position.y
    .m43 = Position.z
  End With
  D3DXVec3TransformCoord vecReturn, LocalVector, matRot
  GetWorldPosition = vecReturn
End Function


